NAME_PROP = "name"
USE_CASE_PROP = "useCase"
STATELESS_PROP = "stateless"
KEEP_IN_CACHE_PROP = "keepInCache"
LAST_FLOW_PROP = "lastUse"

CALLBACK_ERROR_STRATEGY_ARG = "on_key_changed_callback_error_strategy"
KEEP_IN_CACHE_ARG = "keep_in_cache"
STATELESS_ARG = "stateless"
USE_CASE_ARG = "use_case"
SECRET_KEY_TYPE_NAME = "SecretKey"
SECRET_KEY_PAIR_TYPE_NAME = "SecretKeyPair"
SECRET_KEY_DEFINITION_TYPE_NAME = "SecretKeyDefinition"



from __future__ import annotations
from typing import TYPE_CHECKING, Any, Optional, Union
from .secret_key import (
    BaseSecretKeyDefinition,
    InitError,
    SecretKeyFlow,
    SecretKeyState,
    SecretKeyUseCase,
)

if TYPE_CHECKING:
    from .secret_key import (
        KeysStore,
        SecretKeyPair,
        SecretKeyValue,
        StrOrBytes,
        StrOrBytesPair,
    )

ERROR_MSG_FRMT = "original_key_definition type is %s"


class OnChangeKeyDefinition(SecretKeyState):
    __original_key_definition: BaseSecretKeyDefinition
    __key_state: SecretKeyState

    def __init__(self, original_key_definition: BaseSecretKeyDefinition):
        if not isinstance(original_key_definition, BaseSecretKeyDefinition):
            raise OnChangeKeyDefinitionInitError(original_key_definition)
        self.__original_key_definition = original_key_definition
        self.__state = original_key_definition.get_key_state()

    @property
    def name(self) -> str:
        return self.__original_key_definition.name

    @property
    def keys_store(self) -> KeysStore:
        return self.__original_key_definition.keys_store

    def is_stateless(self) -> bool:
        return self.__original_key_definition.is_stateless()

    def is_stated(self) -> bool:
        return not self.__original_key_definition.is_stated()

    @property
    def use_case(self) -> SecretKeyUseCase:
        return self.__original_key_definition.use_case

    def is_keep_in_cache(self) -> bool:
        return self.__original_key_definition.is_keep_in_cache()

    def get_last_flow(self) -> Optional[SecretKeyFlow]:
        return self.__state.get_last_flow()

    def set_last_flow(self, last_use: SecretKeyFlow) -> None:
        self.__state.set_last_flow(last_use)

    def get_previous_keys(self) -> Optional[Union[SecretKeyValue, SecretKeyPair]]:
        return self.__state.get_previous_keys()

    def set_previous_keys(self, keys: Union[StrOrBytes, StrOrBytesPair]) -> None:
        self.__state.set_previous_keys(keys)

    def clean_state(self) -> None:
        self.__state.clean_state()


class OnChangeKeyDefinitionInitError(InitError):
    def __init__(self, original_key_definition: Any) -> None:
        super().__init__(
            "OnChangeKeyDefinition",
            ERROR_MSG_FRMT % str(type(original_key_definition)),
        )



from logging import DEBUG

KEEP_STATE = "keep_state"
KEY = "key"
STATE = "state"
KEYS_STORE = "keys_store"
FORWARD_PATH_FLOW_STATE = "forward"
ENCRYPTED_STATE = "forward"
BACK_PATH_FLOW_STATE = "back"
DECRYPTED_STATE = "back"
ENCRYPTION_KEY_TYPE = "encrypt"
DECRYPTION_KEY_TYPE = "decrypt"
PUBLIC_KEY_TYPE = "public"
PRIVATE_KEY_TYPE = "private"
ON_CHANGES_CALLBACKS = "callback"


DEFINE_KEY_LOG_MESSAGE = 'Defining the key "%s"'
DEFINE_KEY_DEBUG_MESSAGE = "New key was defined %s"


GET_KEY_INFO_MESSAGE = 'requested to get key of "{}"'
GET_KEY_DEBUG_MESSAGE = 'requested to get key of "{}" for {}'
LOG_GEY_DEBUG_MESSAGE = 'rv_key is "%s"'
KEY_CHANGED_DEBUG_MESSAGE = (
    'the key "{}" is changed from {} to {} registered callbacks will be executed'
)
KEY_CHANGED_INFO_MESSAGE = (
    'the key "{}" is changed, registered callbacks will be executed'
)
REGISTER_ON_CHANGE_LOG_MESSAGE = 'registering new OnChange callback for "%s"'

TRACE_LEVEL = DEBUG - 1
TRACE_LEVEL_NAME = "TRACE"



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional, Union

if TYPE_CHECKING:
    from .secret_key import SecretKeyPair, SecretKeyValue
    from .secret_key_use_case import SecretKeyFlow, SecretKeyUseCase
    from .types import StrOrBytes, StrOrBytesPair


class SecretKeyState(ABC):
    @abstractmethod
    def get_last_flow(self) -> Optional[SecretKeyFlow]:
        pass

    @abstractmethod
    def set_last_flow(self, last_use: SecretKeyFlow) -> None:
        pass

    @abstractmethod
    def get_previous_keys(self) -> Optional[Union[SecretKeyValue, SecretKeyPair]]:
        pass

    @abstractmethod
    def set_previous_keys(self, keys: Union[StrOrBytes, StrOrBytesPair]) -> None:
        pass

    @abstractmethod
    def clean_state(self) -> None:
        pass



DEFINE_KEY_LOG_FORMAT = 'Defining the key "%s"'
SUCCESS_DEFINE_KEY_LOG_FORMAT = "New key was defined %s"

GET_KEY_INFO_FORMAT = 'requested to get key of "{}"'
GET_KEY_DEBUG_FORMAT = 'requested to get key of "{}" for {}'
RV_KEY_LOG_FORMAT = 'rv_key is "%s"'
KEY_CHANGED_INFO_FORMAT = (
    'the key "{}" is changed, registered callbacks will be executed'
)
KEY_CHANGED_DEBUG_FORMAT = (
    'the key "{}" is changed from {} to {} registered callbacks will be executed'
)
REGISTER_ON_CHANGE_LOG_FORMAT = 'registering new OnChange callback for "%s"'

ON_SKIP_LOG_FORMAT = 'Skip to next onChange callbacks execution of "{key_name}": Failed to execute {callback_name} - {error}'
ON_HALT_LOG_FORMAT = 'Halt onChange callbacks execution of "{key_name}": Failed to execute {callback_name} - {error}'
CLEAN_PREV_KEYS_LOG_FORMAT = "clean previous '%s' keys from cache"
CLEAN_KEYS_LOG_FORMAT = "clean '%s' keys from cache"



from __future__ import annotations
from enum import Enum
from ..consts import BACK_PATH_FLOW_STATE, FORWARD_PATH_FLOW_STATE
from ..errors import KeysManagementError
from .errors import InvalidFlowNameError


class SecretKeyFlow(Enum):
    DEFAULT = 1
    AUTHENTICATION = 1
    FORWARD_PATH = 2
    BACK_PATH = 3
    ENCRYPTION = 2
    DECRYPTION = 3

    @staticmethod
    def get(str_val: str) -> SecretKeyFlow:
        if str_val == FORWARD_PATH_FLOW_STATE:
            return SecretKeyFlow.FORWARD_PATH
        if str_val == BACK_PATH_FLOW_STATE:
            return SecretKeyFlow.BACK_PATH
        try:
            return SecretKeyFlow[str_val.upper()]
        except KeyError:
            raise InvalidFlowNameError(str_val)


class SecretKeyUseCase(Enum):
    ROUND_TRIP = 1
    ONE_WAY_TRIP = 2
    ENCRYPTION_DECRYPTION = 1



from __future__ import annotations
import inspect
from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Dict, Optional, Union
from unittest.mock import Mock
from ..errors import OnKeyChangedCallbackErrorStrategy
from .consts import (
    CALLBACK_ERROR_STRATEGY_ARG,
    KEEP_IN_CACHE_ARG,
    KEEP_IN_CACHE_PROP,
    LAST_FLOW_PROP,
    NAME_PROP,
    STATELESS_ARG,
    STATELESS_PROP,
    USE_CASE_ARG,
    USE_CASE_PROP,
)
from .errors import InitError, SecretKeyDefinitionInitError
from .key_state import SecretKeyState
from .log_messages_consts import (
    ERROR_STRATEGY_TYPE_ERR_MSG,
    KEEP_IN_CACHE_NOT_BOOL_ERR_MSG,
    NAME_PROPERTY_IS_EMPTY_MSG,
    SHOULD_NOT_CONTAINS_ARGS_MSG,
    STATELESS_NOT_BOOL_ERR_MSG,
    STORE_IS_NOT_CALLABLE_MSG,
    USE_CASE_PROP_TYPE_ERR,
)
from .secret_key import SecretKeyFactory, SecretKeyPair, SecretKeyValue
from .secret_key_use_case import SecretKeyFlow, SecretKeyUseCase

if TYPE_CHECKING:
    from ..key_changed_utils import KeyChangedCallback
    from .types import KeysStore, StrOrBytes, StrOrBytesPair


class BaseSecretKeyDefinition(ABC):
    _name: str
    _store: KeysStore
    _use_case: SecretKeyUseCase
    _stateless: bool
    _keep_in_cache: bool
    _on_key_changed_callback_error_strategy: OnKeyChangedCallbackErrorStrategy
    _on_change_callbacks: Dict[str, KeyChangedCallback]

    def __init__(self, name: str, keys_store: KeysStore, **kwargs: Any):
        self._name = name
        self._keys_store = keys_store
        self._use_case = kwargs.get(
            USE_CASE_ARG, SecretKeyUseCase.ENCRYPTION_DECRYPTION
        )
        self._stateless = kwargs.get(STATELESS_ARG, True)
        self._keep_in_cache = kwargs.get(KEEP_IN_CACHE_ARG, True)
        self._on_key_changed_callback_error_strategy = kwargs.get(
            CALLBACK_ERROR_STRATEGY_ARG,
            OnKeyChangedCallbackErrorStrategy.HALT,
        )
        self._on_change_callbacks = OrderedDict()
        self._validate_properties()

    def _validate_properties(self) -> None:
        if not isinstance(self._name, str) or len(self._name) == 0:
            raise SecretKeyDefinitionInitError(NAME_PROPERTY_IS_EMPTY_MSG)
        if not callable(self._keys_store):
            raise SecretKeyDefinitionInitError(STORE_IS_NOT_CALLABLE_MSG)
        if (
            not isinstance(self._keys_store, Mock)
            and len(inspect.signature(self._keys_store).parameters) > 0
        ):
            raise SecretKeyDefinitionInitError(SHOULD_NOT_CONTAINS_ARGS_MSG)
        if not isinstance(self._use_case, SecretKeyUseCase):
            raise SecretKeyDefinitionInitError(USE_CASE_PROP_TYPE_ERR)
        if not isinstance(self._stateless, bool):
            raise SecretKeyDefinitionInitError(STATELESS_NOT_BOOL_ERR_MSG)
        if not isinstance(self._keep_in_cache, bool):
            raise SecretKeyDefinitionInitError(KEEP_IN_CACHE_NOT_BOOL_ERR_MSG)
        if not isinstance(
            self._on_key_changed_callback_error_strategy,
            OnKeyChangedCallbackErrorStrategy,
        ):
            raise SecretKeyDefinitionInitError(ERROR_STRATEGY_TYPE_ERR_MSG)

    @property
    def name(self) -> str:
        return self._name

    @property
    def keys_store(self) -> KeysStore:
        return self._keys_store

    def is_stateless(self) -> bool:
        return self._stateless

    def is_stated(self) -> bool:
        return not self._stateless

    @property
    def use_case(self) -> SecretKeyUseCase:
        return self._use_case

    @property
    def on_change_callbacks(self) -> Dict[str, KeyChangedCallback]:
        return self._on_change_callbacks

    def is_keep_in_cache(self) -> bool:
        return self._keep_in_cache

    @abstractmethod
    def get_key_state(self) -> SecretKeyState:
        pass

    @abstractmethod
    def set_key_state(self, key_state: SecretKeyState) -> None:
        pass

    @property
    def on_key_changed_callback_error_strategy(
        self,
    ) -> OnKeyChangedCallbackErrorStrategy:
        return self._on_key_changed_callback_error_strategy

    def __str__(self) -> str:
        return str(self.asdict())

    def asdict(self) -> Dict[str, Any]:
        return {
            NAME_PROP: self._name,
            USE_CASE_PROP: self._use_case.name,
            STATELESS_PROP: self._stateless,
            KEEP_IN_CACHE_PROP: self._keep_in_cache,
        }

    def __call__(self) -> Union[StrOrBytes, StrOrBytesPair]:
        return self._keys_store()


KeyContent = Optional[Union[SecretKeyValue, SecretKeyPair]]


class SecretKeyDefinition(BaseSecretKeyDefinition, SecretKeyState):
    _last_use: Optional[SecretKeyFlow]
    _current_keys: KeyContent
    _previous_keys: KeyContent

    def __init__(self, name: str, keys_store: KeysStore, **kwargs: Any) -> None:
        super(SecretKeyDefinition, self).__init__(name, keys_store, **kwargs)
        self._last_use = None
        self._current_keys = None
        self._previous_keys = None

    def set_keys_from_store(self) -> None:
        self._previous_keys = self._current_keys
        self._current_keys = SecretKeyPair(self._keys_store())

    @property
    def keys(self) -> KeyContent:
        return self._current_keys

    @property
    def previous_keys(self) -> KeyContent:
        return self._previous_keys

    def set_previous_keys(
        self,
        keys: Optional[
            Union[StrOrBytes, StrOrBytesPair, SecretKeyValue, SecretKeyPair]
        ],
    ) -> None:
        if isinstance(keys, (SecretKeyValue, SecretKeyPair)):
            self._current_keys = keys
        else:
            self._current_keys = SecretKeyPair(keys)  # type: ignore[arg-type]

    def get_keys_or_previous(self) -> KeyContent:
        if self._current_keys is not None:
            return self._current_keys
        else:
            return self._previous_keys

    def get_previous_or_current_keys(self) -> KeyContent:
        if self._previous_keys is not None:
            return self._previous_keys
        return self._current_keys

    def get_last_flow(self) -> Optional[SecretKeyFlow]:
        return self._last_use

    def set_last_flow(self, use_case: Optional[SecretKeyFlow]) -> None:
        self._last_use = use_case

    def has_keys(self) -> bool:
        return self._current_keys is not None or self._previous_keys is not None

    def clean_keys(self) -> None:
        self._previous_keys = self.get_keys_or_previous()
        self._current_keys = None

    def clean_previous_keys(self) -> None:
        self._previous_keys = None

    def get_key_state(self) -> SecretKeyState:
        return self

    def set_key_state(self, key_state: SecretKeyState) -> None:
        self.set_last_flow(key_state.get_last_flow())
        previous_keys: Optional[
            Union[SecretKeyValue, SecretKeyPair]
        ] = key_state.get_previous_keys()
        if previous_keys != self._previous_keys and previous_keys is not None:
            self.set_previous_keys(previous_keys)

    def clean_state(self) -> None:
        self.clean_keys()
        self.clean_previous_keys()

    def asdict(self) -> Dict[str, Any]:
        rv = super(SecretKeyDefinition, self).asdict()
        rv.update({LAST_FLOW_PROP: str(self._last_use)})
        return rv

    def get_previous_keys(self) -> Optional[Union[SecretKeyValue, SecretKeyPair]]:
        pass



from __future__ import annotations
from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from keys_management.key_changed_utils import KeyChangedContext

FETCH_AND_SET_ERROR_MSG_FORMAT = (
    "Failed to fetch and set key state for {key_name} from_repo: {why}"
)
KEY_IS_NOT_DEFINED_ERROR_MSG_FORMAT = "KeyIsNotDefinedError: key_name is '%s'"
KEY_CHANGED_ERROR_MSG_FORMAT = "KeyChangedError: key_name is '%s'"
GET_KEY_ERROR_MSG_FORMAT = "failed to get key for {key_name}: {reason}"


class KeysManagementError(RuntimeError):
    pass


class OnKeyChangedCallbackErrorStrategy(Enum):
    RAISE_IMMEDIATELY = (1,)
    SKIP_AND_RAISE = (2,)
    SKIP = (3,)
    HALT = (4,)


class GetKeyError(KeysManagementError):
    def __init__(self, key_name: str, reason: str = None) -> None:
        super().__init__(
            GET_KEY_ERROR_MSG_FORMAT.format(
                key_name=key_name, reason=reason if reason else ""
            )
        )


class KeyChangedError(KeysManagementError):
    def __init__(
        self, key_name: str, key_changed_context: KeyChangedContext = None
    ) -> None:
        super().__init__(key_name, key_changed_context)

    def __str__(self) -> str:
        return KEY_CHANGED_ERROR_MSG_FORMAT % self.args[0]


class KeyIsNotDefinedError(KeysManagementError):
    def __init__(self, key_name: str) -> None:
        super().__init__(key_name)

    def __str__(self) -> str:
        return KEY_IS_NOT_DEFINED_ERROR_MSG_FORMAT % self.args[0]


class FetchAndSetStateFromRepoError(KeysManagementError):
    def __init__(self, key_name: str, why: str = None) -> None:
        super().__init__(
            FETCH_AND_SET_ERROR_MSG_FORMAT.format(
                key_name=key_name, why=why if why else ""
            )
        )


class InvalidKeyStateError(FetchAndSetStateFromRepoError):
    def __init__(self, key_name: str, why: str = None) -> None:
        super().__init__(key_name, why)



from typing import Any
from ..errors import KeysManagementError
from .consts import (
    SECRET_KEY_DEFINITION_TYPE_NAME,
    SECRET_KEY_PAIR_TYPE_NAME,
    SECRET_KEY_TYPE_NAME,
)

INIT_ERR_MSG_FRMT = "Failed to init {what_to_init}: {why}"
SECRET_KEY_ERR_MSG_FRMT = "secret_key_value type is %s"
PAIR_INIT_ERR_MSG_FRMT = "secret_key_pair_values type is %s"
FLOW_NAME_ERR_MSG_FRMT = "Failed to parse %s to SecretKeyFlow"


class InitError(KeysManagementError):
    def __init__(self, what_to_init: str, why: str) -> None:
        super().__init__(INIT_ERR_MSG_FRMT.format(what_to_init=what_to_init, why=why))


class SecretKeyInitError(InitError):
    def __init__(self, secret_key_value: Any) -> None:
        super().__init__(
            SECRET_KEY_TYPE_NAME, SECRET_KEY_ERR_MSG_FRMT % str(type(secret_key_value))
        )


class SecretKeyPairInitError(InitError):
    def __init__(self, secret_key_pair_values: Any) -> None:
        super().__init__(
            SECRET_KEY_PAIR_TYPE_NAME,
            PAIR_INIT_ERR_MSG_FRMT % str(type(secret_key_pair_values)),
        )


class InvalidFlowNameError(KeysManagementError):
    def __init__(self, bad_value: str) -> None:
        super().__init__(FLOW_NAME_ERR_MSG_FRMT % bad_value)


class SecretKeyDefinitionInitError(InitError):
    def __init__(self, reason: str) -> None:
        super().__init__(SECRET_KEY_DEFINITION_TYPE_NAME, reason)



from typing import TYPE_CHECKING
from .dependecies import CryptoTool, StateRepoInterface
from .errors import (
    GetKeyError,
    KeyChangedError,
    KeyIsNotDefinedError,
    KeysManagementError,
)
from .key_changed_utils import OnKeyChangedCallbackErrorStrategy
from .keys_management import KeysManagement, KeysManagementImpl
from .on_change_key_definition import (
    OnChangeKeyDefinition,
    OnChangeKeyDefinitionInitError,
)
from .secret_key import (
    BaseSecretKeyDefinition,
    InvalidFlowNameError,
    SecretKeyDefinitionInitError,
    SecretKeyFlow,
    SecretKeyUseCase,
)

if TYPE_CHECKING:
    from .key_changed_utils import KeyChangedCallback
    from .secret_key import KeysStore, StrOrBytes, StrOrBytesPair



from typing import Any, Dict


class StateRepoInterface(object):
    def write_state(self, key: str, key_state: Any) -> None:
        raise NotImplementedError()

    def read_state(self, key: str) -> Dict:
        raise NotImplementedError()


class CryptoTool(object):
    def encrypt(self, data: Any) -> Any:
        raise NotImplementedError()

    def decrypt(self, encrypted_data: Any) -> Any:
        raise NotImplementedError()



from __future__ import annotations
import logging
from typing import TYPE_CHECKING, Callable, Dict, Optional, Union, cast
from .consts import KEY, STATE, TRACE_LEVEL, TRACE_LEVEL_NAME
from .dependecies import CryptoTool, StateRepoInterface
from .errors import (
    FetchAndSetStateFromRepoError,
    GetKeyError,
    InvalidKeyStateError,
    KeyChangedError,
    KeyIsNotDefinedError,
    OnKeyChangedCallbackErrorStrategy,
)
from .key_changed_utils import KeyChangedContext
from .log_messages_consts import (
    CLEAN_KEYS_LOG_FORMAT,
    CLEAN_PREV_KEYS_LOG_FORMAT,
    DEFINE_KEY_LOG_FORMAT,
    GET_KEY_DEBUG_FORMAT,
    GET_KEY_INFO_FORMAT,
    KEY_CHANGED_DEBUG_FORMAT,
    KEY_CHANGED_INFO_FORMAT,
    ON_HALT_LOG_FORMAT,
    ON_SKIP_LOG_FORMAT,
    REGISTER_ON_CHANGE_LOG_FORMAT,
    RV_KEY_LOG_FORMAT,
    SUCCESS_DEFINE_KEY_LOG_FORMAT,
)
from .secret_key import (
    InvalidFlowNameError,
    SecretKeyDefinition,
    SecretKeyFlow,
    SecretKeyPair,
    SecretKeyUseCase,
    SecretKeyValue,
)

if TYPE_CHECKING:
    from .key_changed_utils import KeyChangedCallback
    from .secret_key import KeysStore, StrOrBytes, StrOrBytesPair

logging.addLevelName(TRACE_LEVEL, TRACE_LEVEL_NAME)
logger = logging.getLogger(__name__)

FLOW_IS_NOT_FLOW_TYPE_MSG = 'flow argument is not type of "SecretKeyFlow"'
FLOW_IS_NOT_DEFAULT_MSG = "flow is not SecretKeyFlow.DEFAULT or aliases"
DEFAULT_CALLBACK_NAME_FORMAT = "{}_callback_{}"


class KeysManagement(object):
    """
    This is the main interface class who exposes the library functionalities API.
    """

    def define_key(
        self,
        name: str,
        keys_store: KeysStore,
        use_case: SecretKeyUseCase,
        stateless: bool,
        keep_in_cache: bool,
        on_key_changed_callback_error_strategy: OnKeyChangedCallbackErrorStrategy = None,
    ) -> KeysManagement:
        raise NotImplementedError()

    def get_key(self, key_name: str, flow: SecretKeyFlow) -> StrOrBytes:
        raise NotImplementedError()

    def get_forward_path_key(self, key_name: str) -> StrOrBytes:
        return self.get_key(key_name, SecretKeyFlow.FORWARD_PATH)

    def get_encrypt_key(self, key_name: str) -> StrOrBytes:
        return self.get_forward_path_key(key_name)

    def get_back_path_key(self, key_name: str) -> StrOrBytes:
        return self.get_key(key_name, SecretKeyFlow.BACK_PATH)

    def get_decrypt_key(self, key_name: str) -> StrOrBytes:
        return self.get_back_path_key(key_name)

    def key_changed(
        self,
        key_name: str,
        old_keys: Union[StrOrBytes, StrOrBytesPair],
        new_keys: Union[StrOrBytes, StrOrBytesPair],
    ) -> None:
        raise NotImplementedError()

    def register_on_change(
        self,
        key_name: str,
        on_change_func: KeyChangedCallback,
        callback_id: str = None,
    ) -> None:
        raise NotImplementedError()

    def save_states(self) -> None:
        raise NotImplementedError()

    def save_state(self, key_name: str) -> None:
        raise NotImplementedError


class KeysManagementImpl(KeysManagement):
    _state_repo: StateRepoInterface
    _crypto_tool: CryptoTool
    _keys_definitions: Dict[str, SecretKeyDefinition]
    _callbacks_executions_error_handling: Dict[
        OnKeyChangedCallbackErrorStrategy, Callable
    ]

    def __init__(
        self,
        state_repo: Optional[StateRepoInterface] = None,
        crypto_tool: Optional[CryptoTool] = None,
    ):
        self._state_repo = (
            state_repo if state_repo is not None else StateRepoInterface()
        )
        self._crypto_tool = crypto_tool if crypto_tool is not None else CryptoTool()
        self._keys_definitions = {}
        self._callbacks_executions_error_handling = {
            OnKeyChangedCallbackErrorStrategy.HALT: KeysManagementImpl._on_halt_strategy,
            OnKeyChangedCallbackErrorStrategy.SKIP: KeysManagementImpl._on_skip_strategy,
            OnKeyChangedCallbackErrorStrategy.SKIP_AND_RAISE: KeysManagementImpl._on_skip_and_raise_strategy,
            OnKeyChangedCallbackErrorStrategy.RAISE_IMMEDIATELY: KeysManagementImpl._on_raise_strategy,
        }

    def define_key(
        self,
        name: str,
        keys_store: KeysStore,
        use_case: SecretKeyUseCase,
        stateless: bool = None,
        keep_in_cache: bool = None,
        on_key_changed_callback_error_strategy: OnKeyChangedCallbackErrorStrategy = None,
    ) -> KeysManagement:
        on_key_changed_callback_error_strategy = (
            on_key_changed_callback_error_strategy
            if on_key_changed_callback_error_strategy is not None
            else OnKeyChangedCallbackErrorStrategy.HALT
        )
        logger.info(DEFINE_KEY_LOG_FORMAT % name)
        key_definition = SecretKeyDefinition(
            name,
            keys_store,
            use_case=use_case,
            stateless=stateless,
            keep_in_cache=keep_in_cache,
            on_key_changed_callback_error_strategy=on_key_changed_callback_error_strategy,
        )
        logger.debug(SUCCESS_DEFINE_KEY_LOG_FORMAT % str(key_definition))
        self._keys_definitions[name] = key_definition
        return self

    def get_key(self, key_name: str, flow: SecretKeyFlow = None) -> StrOrBytes:
        try:
            if not logger.isEnabledFor(logging.DEBUG):
                logger.info(GET_KEY_INFO_FORMAT.format(key_name))
            self._validate_key_name(key_name)
            key_definition = self._keys_definitions[key_name]
            flow = self._determine_flow(key_definition) if flow is None else flow
            logger.debug(GET_KEY_DEBUG_FORMAT.format(key_name, flow.name))
            rv_key = self._get_key_by_flow(key_definition, flow)
            logger.debug(RV_KEY_LOG_FORMAT, str(rv_key))
            self._update_key_definition_state(key_definition, flow)
            return rv_key.get_value()
        except GetKeyError as e:
            raise e
        except Exception as e:
            raise GetKeyError(key_name) from e

    def _validate_key_name(self, key_name: str) -> None:
        logger.info(REGISTER_ON_CHANGE_LOG_FORMAT % key_name)
        if key_name not in self._keys_definitions:
            raise KeyIsNotDefinedError(key_name)

    def _determine_flow(self, key_definition: SecretKeyDefinition) -> SecretKeyFlow:
        if key_definition.use_case not in {SecretKeyUseCase.ROUND_TRIP, None}:
            return SecretKeyFlow.DEFAULT
        else:
            return self.__determine_flow_by_previous_flow(key_definition)

    def __determine_flow_by_previous_flow(
        self, key_definition: SecretKeyDefinition
    ) -> SecretKeyFlow:
        prev_flow = self.__get_previous_flow(key_definition)
        if prev_flow == SecretKeyFlow.FORWARD_PATH:
            return SecretKeyFlow.BACK_PATH
        else:
            return SecretKeyFlow.FORWARD_PATH

    def __get_previous_flow(
        self, key_definition: SecretKeyDefinition
    ) -> Optional[SecretKeyFlow]:
        prev_flow = key_definition.get_last_flow()
        if prev_flow is None and key_definition.is_stated():
            self._fetch_and_set_state_from_repo(key_definition)
            prev_flow = key_definition.get_last_flow()
        return prev_flow

    def _fetch_and_set_state_from_repo(
        self, key_definition: SecretKeyDefinition
    ) -> None:
        try:
            raw_state = self._crypto_tool.decrypt(
                self._state_repo.read_state(key_definition.name)
            )
            key_definition.set_last_flow(SecretKeyFlow.get(raw_state[STATE]))
            if KEY in raw_state:
                key_definition.set_previous_keys(raw_state[KEY])
        except InvalidFlowNameError as e:
            raise InvalidKeyStateError(key_definition.name) from e
        except Exception as e:
            raise FetchAndSetStateFromRepoError(key_definition.name) from e

    def _get_key_by_flow(
        self,
        key_definition: SecretKeyDefinition,
        flow: SecretKeyFlow,
    ) -> SecretKeyValue:
        if key_definition.use_case == SecretKeyUseCase.ROUND_TRIP:
            return self._get_key_round_trip_case(key_definition, flow)
        else:
            return self._get_key_one_way_case(key_definition, flow)

    def _get_key_round_trip_case(
        self,
        key_definition: SecretKeyDefinition,
        flow: SecretKeyFlow,
    ) -> SecretKeyValue:
        if flow == SecretKeyFlow.FORWARD_PATH:
            return self._get_key_for_forward(key_definition)
        else:  # flow == SecretKeyFlow.BACK_PATH:
            return self._get_key_for_back_path(key_definition)

    def _get_key_for_forward(
        self, key_definition: SecretKeyDefinition
    ) -> SecretKeyValue:
        key_definition.set_keys_from_store()
        return key_definition.keys.forward_key  # type: ignore[union-attr]

    def _get_key_for_back_path(
        self, key_definition: SecretKeyDefinition
    ) -> SecretKeyValue:
        if not key_definition.has_keys():
            if key_definition.get_last_flow() is None and key_definition.is_stated():
                self._fetch_and_set_state_from_repo(key_definition)
                if not key_definition.has_keys():
                    key_definition.set_keys_from_store()
            else:
                key_definition.set_keys_from_store()
        return key_definition.get_previous_or_current_keys().back_path_key  # type: ignore[union-attr]

    def _get_key_one_way_case(
        self,
        key_definition: SecretKeyDefinition,
        flow: SecretKeyFlow,
    ) -> SecretKeyValue:
        if flow != SecretKeyFlow.DEFAULT:
            GetKeyError(
                key_name=key_definition.name,
                reason=FLOW_IS_NOT_DEFAULT_MSG,
            )
        return SecretKeyValue(cast(Union[str, bytes], key_definition.keys_store()))

    def _update_key_definition_state(
        self, key_definition: SecretKeyDefinition, flow: SecretKeyFlow
    ) -> None:
        # todo test it
        logger.log(TRACE_LEVEL, CLEAN_KEYS_LOG_FORMAT % key_definition.name)
        key_definition.clean_keys()
        key_definition.set_last_flow(flow)
        if self._is_clean_previous_keys(key_definition, flow):
            logger.log(TRACE_LEVEL, CLEAN_PREV_KEYS_LOG_FORMAT % key_definition.name)
            key_definition.clean_previous_keys()

    @staticmethod
    def _is_clean_previous_keys(
        key_definition: SecretKeyDefinition,
        current_flow: SecretKeyFlow,
    ) -> bool:
        return current_flow == SecretKeyFlow.DEFAULT or (
            current_flow == SecretKeyFlow.BACK_PATH
            and not key_definition.is_keep_in_cache()
        )

    def key_changed(
        self,
        key_name: str,
        old_keys: Union[StrOrBytes, StrOrBytesPair] = None,
        new_keys: Union[StrOrBytes, StrOrBytesPair] = None,
    ) -> None:
        try:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(
                    KEY_CHANGED_DEBUG_FORMAT.format(
                        key_name,
                        str(SecretKeyPair(old_keys)) if old_keys is not None else None,
                        str(SecretKeyPair(new_keys)) if new_keys is not None else None,
                    )
                )
            else:
                logger.info(KEY_CHANGED_INFO_FORMAT.format(key_name))
            self._validate_key_name(key_name)
            key_changed_context: KeyChangedContext = self._create_context(
                self._keys_definitions[key_name], old_keys, new_keys
            )
            key_changed_context.run_callbacks()
        except KeyIsNotDefinedError as e:
            raise KeyChangedError(key_name) from e
        except DoNothing:
            pass

    @staticmethod
    def _on_halt_strategy(
        key_name: str,
        callback_name: str,
        key_changed_context: KeyChangedContext,
    ) -> None:
        logger.error(
            ON_HALT_LOG_FORMAT.format(
                key_name=key_name,
                callback_name=callback_name,
                error=key_changed_context[callback_name]["error"],
            )
        )
        raise DoNothing()

    @staticmethod
    def _on_skip_strategy(
        key_name: str,
        callback_name: str,
        key_changed_context: KeyChangedContext,
    ) -> None:
        logger.error(
            ON_SKIP_LOG_FORMAT.format(
                key_name=key_name,
                callback_name=callback_name,
                error=key_changed_context[callback_name]["error"],
            )
        )

    @staticmethod
    def _on_skip_and_raise_strategy(
        key_name: str,
        callback_name: str,
        key_changed_context: KeyChangedContext,
    ) -> None:
        pass

    @staticmethod
    def _on_raise_strategy(
        key_name: str,
        callback_name: str,
        key_changed_context: KeyChangedContext,
    ) -> None:
        raise KeyChangedError(key_name, key_changed_context)

    def _create_context(
        self,
        key_definition: SecretKeyDefinition,
        old_keys: Optional[Union[StrOrBytes, StrOrBytesPair]],
        new_keys: Optional[Union[StrOrBytes, StrOrBytesPair]],
    ) -> KeyChangedContext:
        return KeyChangedContext(
            key_definition,
            self._callbacks_executions_error_handling[
                key_definition.on_key_changed_callback_error_strategy
            ],
            old_keys,
            new_keys,
        )

    def register_on_change(
        self,
        key_name: str,
        on_change_func: KeyChangedCallback,
        callback_name: str = None,
    ) -> None:
        self._validate_key_name(key_name)
        callbacks = self._keys_definitions[key_name].on_change_callbacks
        callback_name = (
            callback_name
            if callback_name is not None
            else DEFAULT_CALLBACK_NAME_FORMAT.format(key_name, str(len(callbacks) + 1))
        )
        callbacks[callback_name] = on_change_func

    def save_state(self, key_name: str) -> None:
        self._validate_key_name(key_name)
        last_flow = self._keys_definitions[key_name].get_last_flow()
        raw_state = {STATE: last_flow}
        if last_flow == SecretKeyFlow.FORWARD_PATH:
            key = self.get_back_path_key(key_name)
            if key is not None:
                raw_state[KEY] = key  # type: ignore[assignment]
        self._write_state(key_name, raw_state)

    def _write_state(self, key_name: str, state: Dict) -> None:
        self._state_repo.write_state(key_name, self._crypto_tool.encrypt(state))

    def save_states(self) -> None:
        for key_name, key_definition in self._keys_definitions.items():
            if key_definition.is_stated():
                self.save_state(key_name)


class DoNothing(Exception):
    pass



from __future__ import annotations
import logging
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Union
from .errors import KeyChangedError, OnKeyChangedCallbackErrorStrategy
from .on_change_key_definition import OnChangeKeyDefinition

logger = logging.getLogger(__name__)


class CallbackStatus(Enum):
    PENDING = 0
    IN_PROGRESS = 1
    FAILED = 2
    SUCCEEDED = 3


if TYPE_CHECKING:
    from .secret_key import StrOrBytes, StrOrBytesPair
    from .secret_key.key_definition import SecretKeyDefinition

    OldKeys = Union[StrOrBytes, StrOrBytesPair]
    newKeys = Union[StrOrBytes, StrOrBytesPair]
    KeyChangedCallback = Callable[
        [
            OldKeys,
            newKeys,
            OnChangeKeyDefinition,
        ],
        None,
    ]
    Callbacks = Dict[str, Dict[str, Any]]


class KeyChangedContext:
    _key_name: str
    _on_change_key_definition: OnChangeKeyDefinition
    _on_key_changed_callback_error_strategy: OnKeyChangedCallbackErrorStrategy
    _strategy_function: Callable[..., None]
    _callbacks: Callbacks
    _has_error: bool
    _old_keys: Optional[Union[StrOrBytes, StrOrBytesPair]]
    _new_keys: Optional[Union[StrOrBytes, StrOrBytesPair]]

    def __init__(
        self,
        key_definition: SecretKeyDefinition,
        on_error_strategy: Callable[..., None],
        old_keys: Optional[Union[StrOrBytes, StrOrBytesPair]],
        new_keys: Optional[Union[StrOrBytes, StrOrBytesPair]],
    ) -> None:
        self._key_name = key_definition.name
        self._on_change_key_definition = OnChangeKeyDefinition(key_definition)
        self._on_key_changed_callback_error_strategy = (
            key_definition.on_key_changed_callback_error_strategy
        )
        self._strategy_function = on_error_strategy  # type: ignore[assignment]
        self._callbacks = self._create_callbacks(key_definition.on_change_callbacks)
        self._old_keys = old_keys
        self._new_keys = new_keys
        self._has_error = False

    def __getitem__(self, item: str) -> Any:
        return self._callbacks[item]

    @staticmethod
    def _create_callbacks(
        on_change_callbacks: Dict[str, KeyChangedCallback]
    ) -> Callbacks:
        return {
            callback_name: {
                "name": callback_name,
                "callback": callback,
                "status": CallbackStatus.PENDING,
            }
            for callback_name, callback in on_change_callbacks.items()
        }

    def run_callbacks(self) -> None:
        for callback_name, callback_ctx in self._callbacks.items():
            try:
                logger.info('Going to execute the callback "%s"' % callback_name)
                callback_ctx["status"] = CallbackStatus.IN_PROGRESS
                callback_ctx["callback"](
                    self._old_keys,
                    self._new_keys,
                    self._on_change_key_definition,
                )
                callback_ctx["status"] = CallbackStatus.SUCCEEDED
            except Exception as e:
                self._has_error = True
                callback_ctx["status"] = CallbackStatus.FAILED
                callback_ctx["error"] = e
                self._strategy_function(self._key_name, callback_name, self)
        if (
            self._on_key_changed_callback_error_strategy
            == OnKeyChangedCallbackErrorStrategy.SKIP_AND_RAISE
            and self._has_error
        ):
            raise KeyChangedError(self._key_name, self)



from __future__ import annotations
from math import floor
from typing import TYPE_CHECKING, Any, Optional, Union
from .errors import InitError, SecretKeyInitError, SecretKeyPairInitError

if TYPE_CHECKING:
    from .types import StrOrBytes, StrOrBytesPair


class SecretKeyValue:
    _val: StrOrBytes

    def __init__(self, secret_key_value: StrOrBytes):
        if isinstance(secret_key_value, (str, bytes)):
            self._val = secret_key_value
        else:
            raise SecretKeyInitError(secret_key_value)

    def __str__(self) -> str:
        return self.censor_key()

    def get_value(self) -> StrOrBytes:
        return self._val

    def censor_key(self) -> str:
        """
        censorString('') = ''; censorString('a') = '*'; censorString('aa') = '**';
        censorString('aaa') = '***'; censorString('aaaa') = 'a**a'; censorString('aaaaa') = 'a***a';
        censorString('aaaaaaa') = 'a****a'; censorString('aaaaaaa') = 'a*****a';
        censorString('aaaaaaaa') = 'aa****aa';
        """
        str_to_censor = str(self._val)
        key_length = len(str_to_censor)

        revealedPartSize = min(4, floor(key_length / 4))
        return (
            str_to_censor[0:revealedPartSize]
            + "*" * (key_length - 2 * revealedPartSize)
            + str_to_censor[key_length - revealedPartSize :]
        )


class SecretKeyPair:
    _forward_path_key: SecretKeyValue
    _back_path_key: SecretKeyValue
    _is_symmetric: bool

    def __init__(self, value_or_values: Union[StrOrBytes, StrOrBytesPair]):
        forward_path_key: StrOrBytes
        back_path_key: StrOrBytes
        if isinstance(value_or_values, (str, bytes)):
            forward_path_key = value_or_values
            back_path_key = value_or_values
        elif isinstance(value_or_values, tuple) and len(value_or_values) == 2:
            forward_path_key, back_path_key = value_or_values
        else:
            raise SecretKeyPairInitError(value_or_values)
        self._forward_path_key = SecretKeyValue(forward_path_key)
        self._back_path_key = SecretKeyValue(back_path_key)
        self._is_symmetric = forward_path_key == back_path_key

    def __str__(self) -> str:
        if self.is_symmetric():
            return '"%s"' % str(self._back_path_key)
        else:
            return 'encrypt: "{}", decrypt: "{}"'.format(
                str(self._forward_path_key), str(self._back_path_key)
            )

    def is_symmetric(self) -> bool:
        return self._is_symmetric

    def is_asymmetric(self) -> bool:
        return not self._is_symmetric

    @property
    def back_path_key(self) -> SecretKeyValue:
        return self._back_path_key

    @property
    def forward_key(self) -> SecretKeyValue:
        return self._forward_path_key


class SecretKeyFactory:
    @staticmethod
    def create(
        value_or_values: Optional[Union[StrOrBytes, StrOrBytesPair]]
    ) -> Optional[Union[SecretKeyValue, SecretKeyPair]]:
        if value_or_values is None:
            return None
        if isinstance(value_or_values, (str, bytes)):
            return SecretKeyValue(value_or_values)
        elif isinstance(value_or_values, tuple):
            return SecretKeyPair(value_or_values)
        raise SecretKeyPairInitError(value_or_values)

    def __call__(
        self, value_or_values: Union[StrOrBytes, StrOrBytesPair]
    ) -> Optional[Union[SecretKeyValue, SecretKeyPair]]:
        return self.create(value_or_values)



ERROR_STRATEGY_TYPE_ERR_MSG = '"on_key_changed_callback_error_strategy" property is not type of "OnKeyChangedCallbackErrorStrategy"'
KEEP_IN_CACHE_NOT_BOOL_ERR_MSG = '"keep_in_cache" property is not boolean'
STATELESS_NOT_BOOL_ERR_MSG = '"stateless" property is not boolean'
USE_CASE_PROP_TYPE_ERR = '"use_case" property is not type of "SecretKeyUseCase"'
SHOULD_NOT_CONTAINS_ARGS_MSG = '"keys_store" signature should not contains args'
STORE_IS_NOT_CALLABLE_MSG = '"keys_store" is not callable'
NAME_PROPERTY_IS_EMPTY_MSG = '"name" property is empty'



from __future__ import annotations
from typing import TYPE_CHECKING
from .errors import InitError, InvalidFlowNameError, SecretKeyDefinitionInitError
from .key_definition import BaseSecretKeyDefinition, SecretKeyDefinition
from .key_state import SecretKeyState
from .secret_key import SecretKeyFactory, SecretKeyPair, SecretKeyValue
from .secret_key_use_case import SecretKeyFlow, SecretKeyUseCase

if TYPE_CHECKING:
    from .types import KeysStore, StrOrBytes, StrOrBytesPair
